@@ -21,7 +21,7 @@
  * Copyright (C) 2001-2002, David A. Parker <david@neongoat.com>.
  * http://libdbi.sourceforge.net
  * 
- * $Id: dbd_pgsql.c,v 1.38 2004/01/04 00:43:20 mhoenicka Exp $
+ * $Id: dbd_pgsql.c,v 1.51 2006/09/21 19:44:13 mhoenicka Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -38,9 +38,16 @@ long long atoll(const char *str);
 long long strtoll(const char *nptr, char **endptr, int base);
 #endif
 
+/* In 7.4 PQfreeNotify was deprecated and PQfreemem is used instead.  A
+   macro exists in 7.4 for backwards compatibility. */
+#ifndef PQfreeNotify   /* must be earlier than 7.4 */
+#define PQfreemem PQfreeNotify
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h> /* for isdigit() */
 
 #include <dbi/dbi.h>
 #include <dbi/dbi-dev.h>
@@ -61,9 +68,9 @@ static const dbi_info_t driver_info = {
 static const char *custom_functions[] = {NULL}; // TODO
 static const char *reserved_words[] = PGSQL_RESERVED_WORDS;
 
-/* encoding strings, array is terminated by empty string */
+/* encoding strings, array is terminated by a pair of empty strings */
 static const char pgsql_encoding_hash[][16] = {
-  /* from , to */
+  /* PostgreSQL , www.iana.org */
   "SQL_ASCII", "US-ASCII",
   "EUC_JP", "EUC-JP",
   "EUC_KR", "EUC-KR",
@@ -85,13 +92,14 @@ static const char pgsql_encoding_hash[][16] = {
   "KOI8", "KOI8-R",
   "WIN", "windows-1251",
   "ALT", "IBM866",
-  ""
+  "", ""
 };
 
 /* forward declarations of internal functions */
 void _translate_postgresql_type(unsigned int oid, unsigned short *type, unsigned int *attribs);
 void _get_field_info(dbi_result_t *result);
 void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned long long rowidx);
+int _dbd_real_connect(dbi_conn_t *conn, const char *db);
 
 /* this function is available through the PostgreSQL client library, but it
    is not declared in any of their headers. I hope this won't break anything */
@@ -116,10 +124,15 @@ int dbd_initialize(dbi_driver_t *driver) {
 }
 
 int dbd_connect(dbi_conn_t *conn) {
+  return _dbd_real_connect(conn, NULL);
+}
+
+int _dbd_real_connect(dbi_conn_t *conn, const char *db) {
 	const char *host = dbi_conn_get_option(conn, "host");
 	const char *username = dbi_conn_get_option(conn, "username");
 	const char *password = dbi_conn_get_option(conn, "password");
-	const char *dbname = dbi_conn_get_option(conn, "dbname");
+	const char *dbname;
+	const char *encoding = dbi_conn_get_option(conn, "encoding");
 	int port = dbi_conn_get_option_numeric(conn, "port");
 
 	/* pgsql specific options */
@@ -131,8 +144,19 @@ int dbd_connect(dbi_conn_t *conn) {
 	char *conninfo;
 	char *conninfo_kludge;
 
-	if (port > 0) asprintf(&port_str, "%d", port);
-	else port_str = NULL;
+	if (db && *db) {
+	  dbname = db;
+	}
+	else {
+	  dbname = dbi_conn_get_option(conn, "dbname");
+	}
+
+	if (port > 0) {
+	  asprintf(&port_str, "%d", port);
+	}
+	else {
+	  port_str = NULL;
+	}
 
 	/* YUCK YUCK YUCK YUCK YUCK. stupid libpq. */
 	if (host && port_str) asprintf(&conninfo_kludge, "host='%s' port='%s'", host, port_str);
@@ -168,6 +192,17 @@ int dbd_connect(dbi_conn_t *conn) {
 		if (dbname) conn->current_db = strdup(dbname);
 	}
 	
+	if (encoding && *encoding) {
+	  /* set connection encoding */
+	  if (strcmp(encoding, "auto")) {
+	    if (PQsetClientEncoding(pgconn, dbd_encoding_from_iana(encoding))) {
+/* 	      printf("could not set client encoding to %s\n", dbd_encoding_from_iana(encoding)); */
+	    }
+	  }
+	  /* else: by default, pgsql uses the database encoding
+	     as the client encoding, nothing to do */
+	}
+
 	return 0;
 }
 
@@ -176,16 +211,16 @@ int dbd_disconnect(dbi_conn_t *conn) {
 	return 0;
 }
 
-int dbd_fetch_row(dbi_result_t *result, unsigned long long rownum) {
+int dbd_fetch_row(dbi_result_t *result, unsigned long long rowidx) {
 	dbi_row_t *row = NULL;
 
-	if (result->result_state == NOTHING_RETURNED) return -1;
+	if (result->result_state == NOTHING_RETURNED) return 0;
 	
 	if (result->result_state == ROWS_RETURNED) {
 		/* get row here */
 		row = _dbd_row_allocate(result->numfields);
-		_get_row_data(result, row, rownum);
-		_dbd_row_finalize(result, row, rownum);
+		_get_row_data(result, row, rowidx);
+		_dbd_row_finalize(result, row, rowidx);
 	}
 	
 	return 1; /* 0 on error, 1 on successful fetchrow */
@@ -196,7 +231,7 @@ int dbd_free_query(dbi_result_t *result) {
 	return 0;
 }
 
-int dbd_goto_row(dbi_result_t *result, unsigned long long row) {
+int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx) {
 	/* libpq doesn't have to do anything, the row index is specified when
 	 * fetching fields */
 	return 1;
@@ -212,40 +247,130 @@ int dbd_get_socket(dbi_conn_t *conn)
 }
 
 const char *dbd_get_encoding(dbi_conn_t *conn){
-
-        char* my_enc;
+	char* my_enc;
+	int n_encoding;
+	const char* encodingopt;
+	char* sql_cmd;
+	dbi_result dbires = NULL;
 	PGconn *pgconn = (PGconn*) conn->connection;
 	
 	if(!pgconn) return NULL;
 
-	/* this is somewhat murky as the pg_encoding_to_char()
-	 function is not declared properly by the PostgreSQL client
-	 library headers.  This may indicate that it is not supposed
-	 to be exported or that it may disappear without a trace
-	 eventually. If it breaks, use a query "SHOW CLIENT_ENCODING"
-	 instead */
-        my_enc = pg_encoding_to_char(PQclientEncoding(pgconn));
+	encodingopt = dbi_conn_get_option(conn, "encoding");
+	if (encodingopt && !strcmp(encodingopt, "auto")) {
+
+	  /* this is somewhat murky as the pg_encoding_to_char()
+	     function is not declared properly by the PostgreSQL client
+	     library headers.  This may indicate that it is not supposed
+	     to be exported or that it may disappear without a trace
+	     eventually. If it breaks, use a query "SHOW CLIENT_ENCODING"
+	     instead */
+	  my_enc = pg_encoding_to_char(PQclientEncoding(pgconn));
+/*  	  printf("use PQclientEncoding, auto\n"); */
+	}
+	else if (encodingopt) {
+	  my_enc = pg_encoding_to_char(PQclientEncoding(pgconn));
+/*  	  printf("use PQclientEncoding, %s\n", encodingopt); */
+	}
+	else {
+	  asprintf(&sql_cmd, "SELECT encoding FROM pg_database WHERE datname='%s'", conn->current_db);
+	  
+	  dbires = dbi_conn_query(conn, sql_cmd);
+	  free(sql_cmd);
+
+	  if (dbires && dbi_result_next_row(dbires)) {
+	    n_encoding = dbi_result_get_int_idx(dbires, 1);
+	    my_enc = pg_encoding_to_char(n_encoding);
+/*  	    printf("select returned encoding %d<<%s\n", n_encoding, my_enc); */
+	  }
+	}
 
 	if (!my_enc) {
 	  return NULL;
 	}
 	else {
-	  int i = 0;
-
-	  /* loop over all even entries in hash and compare to my_enc */
-	  while (*pgsql_encoding_hash[i]) {
-	    if (!strcmp(pgsql_encoding_hash[i], my_enc)) {
-	      /* return corresponding odd entry */
-	      return pgsql_encoding_hash[i+1];
-	    }
-	    i+=2;
-	  }
-
-	  /* don't know how to translate, return original string */
-	  return my_enc;
+	  return dbd_encoding_to_iana(my_enc);
 	}
 }
 
+const char* dbd_encoding_to_iana(const char *db_encoding) {
+  int i = 0;
+
+  /* loop over all even entries in hash and compare to penc */
+  while (*pgsql_encoding_hash[i]) {
+    if (!strcmp(pgsql_encoding_hash[i], db_encoding)) {
+      /* return corresponding odd entry */
+      return pgsql_encoding_hash[i+1];
+    }
+    i+=2;
+  }
+
+  /* don't know how to translate, return original encoding */
+  return db_encoding;
+}
+
+const char* dbd_encoding_from_iana(const char *iana_encoding) {
+  int i = 0;
+
+  /* loop over all odd entries in hash and compare to ienc */
+  while (*pgsql_encoding_hash[i+1]) {
+    if (!strcmp(pgsql_encoding_hash[i+1], iana_encoding)) {
+      /* return corresponding even entry */
+      return pgsql_encoding_hash[i];
+    }
+    i+=2;
+  }
+
+  /* don't know how to translate, return original encoding */
+  return iana_encoding;
+}
+
+char *dbd_get_engine_version(dbi_conn_t *conn, char *versionstring) {
+  dbi_result_t *dbi_result;
+  const char *versioninfo = NULL;
+
+  /* initialize return string */
+  *versionstring = '\0';
+
+  dbi_result = dbd_query(conn, "SELECT VERSION()");
+
+  /* this query will return something like:
+     PostgreSQL 8.0.1 on i386-portbld-freebsd5.4, compiled by GCC cc (GCC) 3.4.2 [FreeBSD] 20040728
+  */
+  if (dbi_result) {
+    if (dbi_result_next_row(dbi_result)) {
+      char *dot = NULL;
+      char *start = NULL;
+      char *stop = NULL;
+      versioninfo = dbi_result_get_string_idx(dbi_result, 1);
+
+      /* try to locate the version number. Look for the first dot, go
+	 back where the number before the dot starts, then walk
+	 forward to the last dot or number */
+      dot = strchr(versioninfo, (int)'.');
+      if (dot) {
+	start = dot-1;
+	while (start>versioninfo && isdigit((int)(*(start-1)))) {
+	  start--;
+	}
+
+	stop = start;
+	while (*(stop+1) && (isdigit((int)(*(stop+1))) || *(stop+1)=='.')) {
+	  stop++;
+	}
+
+	if (stop-start < VERSIONSTRING_LENGTH) {
+	  strncpy(versionstring, start, stop-start+1);
+	  versionstring[stop-start+1] = '\0';
+	}
+      }
+    }
+    dbi_result_free(dbi_result);
+  }
+
+  return versionstring;
+}
+
 dbi_result_t *dbd_list_dbs(dbi_conn_t *conn, const char *pattern) {
 	dbi_result_t *res;
 	char *sql_cmd;
@@ -274,9 +399,9 @@ dbi_result_t *dbd_list_tables(dbi_conn_t *conn, const char *db, const char *patt
 	}
 }
 
-int dbd_quote_string(dbi_driver_t *driver, const char *orig, char *dest) {
+size_t dbd_quote_string(dbi_driver_t *driver, const char *orig, char *dest) {
 	/* foo's -> 'foo\'s' */
-	int len;
+	size_t len;
 
 	strcpy(dest, "'");
 	len = PQescapeString(dest+1, orig, strlen(orig));
@@ -285,6 +410,36 @@ int dbd_quote_string(dbi_driver_t *driver, const char *orig, char *dest) {
 	return len+2;
 }
 
+size_t dbd_conn_quote_string(dbi_conn_t *conn, const char *orig, char *dest) {
+  return dbd_quote_string(conn->driver, orig, dest);
+}
+
+size_t dbd_quote_binary(dbi_conn_t *conn, const unsigned char* orig, size_t from_length, unsigned char **ptr_dest) {
+  unsigned char *temp = NULL;
+  unsigned char *quoted_temp = NULL;
+  size_t to_length;
+
+  temp = PQescapeBytea(orig, from_length, &to_length);
+
+  if (!temp) {
+    return 0;
+  }
+
+  if ((quoted_temp = malloc(to_length+2)) == NULL) {
+    PQfreemem((void *)temp);
+    return 0;
+  }
+  
+  strcpy((char *)quoted_temp, "'");
+  strcpy((char *)(quoted_temp+1), (char *)temp);
+  strcat((char *)quoted_temp, "'");
+
+  PQfreemem((void*)temp);
+
+  *ptr_dest = quoted_temp;
+  return to_length;
+}
+
 dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
 	/* allocate a new dbi_result_t and fill its applicable members:
 	 * 
@@ -302,18 +457,18 @@ dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
 		return NULL;
 	}
 
-	result = _dbd_result_create(conn, (void *)res, PQntuples(res), atol(PQcmdTuples(res)));
-	_dbd_result_set_numfields(result, PQnfields((PGresult *)result->result_handle));
+	result = _dbd_result_create(conn, (void *)res, (unsigned long long)PQntuples(res), (unsigned long long)atoll(PQcmdTuples(res)));
+	_dbd_result_set_numfields(result, (unsigned int)PQnfields((PGresult *)result->result_handle));
 	_get_field_info(result);
 
 	return result;
 }
 
-dbi_result_t *dbd_query_null(dbi_conn_t *conn, const unsigned char *statement, unsigned long st_length) {
+dbi_result_t *dbd_query_null(dbi_conn_t *conn, const unsigned char *statement, size_t st_length) {
 	return NULL;
 }
 
-char *dbd_select_db(dbi_conn_t *conn, const char *db) {
+const char *dbd_select_db(dbi_conn_t *conn, const char *db) {
   /* postgresql doesn't support switching databases without reconnecting */
   if (!db || !*db) {
     return NULL;
@@ -324,12 +479,11 @@ char *dbd_select_db(dbi_conn_t *conn, const char *db) {
     conn->connection = NULL;
   }
 
-  dbi_conn_set_option(conn, "dbname", db);
-  if (dbd_connect(conn)) {
+  if (_dbd_real_connect(conn, db)) {
     return NULL;
   }
 
-  return (char *)db;
+  return db;
 }
 
 int dbd_geterror(dbi_conn_t *conn, int *errno, char **errstr) {
@@ -356,7 +510,7 @@ unsigned long long dbd_get_seq_last(dbi_conn_t *conn, const char *sequence) {
 	if (result) {
 		rawdata = PQgetvalue((PGresult *)result->result_handle, 0, 0);
 		if (rawdata) {
-			seq_last = atoll(rawdata);
+			seq_last = (unsigned long long)atoll(rawdata);
 		}
 		dbi_result_free((dbi_result)result);
 	}
@@ -378,7 +532,7 @@ unsigned long long dbd_get_seq_next(dbi_conn_t *conn, const char *sequence) {
 	if (result) {	
 		rawdata = PQgetvalue((PGresult *)result->result_handle, 0, 0);
 		if (rawdata) {
-			seq_next = atoll(rawdata);
+			seq_next = (unsigned long long)atoll(rawdata);
 		}
 		dbi_result_free((dbi_result)result);
 	}
@@ -388,8 +542,12 @@ unsigned long long dbd_get_seq_next(dbi_conn_t *conn, const char *sequence) {
 
 int dbd_ping(dbi_conn_t *conn) {
 	PGconn *pgsql = (PGconn *)conn->connection;
+	PGresult *res;
 
-	PQexec(pgsql, "SELECT 1");
+	res = PQexec(pgsql, "SELECT 1");
+	if (res) {
+	  PQclear (res);
+	}
 
 	if (PQstatus(pgsql) == CONNECTION_OK) {
 		return 1;
@@ -410,6 +568,7 @@ void _translate_postgresql_type(unsigned int oid, unsigned short *type, unsigned
 	unsigned int _type = 0;
 	unsigned int _attribs = 0;
 
+/* 	  fprintf(stderr, "oid went to %d\n", oid); */
 	switch (oid) {
 		case PG_TYPE_CHAR:
 			_type = DBI_TYPE_INTEGER;
@@ -441,20 +600,23 @@ void _translate_postgresql_type(unsigned int oid, unsigned short *type, unsigned
 			_type = DBI_TYPE_DECIMAL;
 			_attribs |= DBI_DECIMAL_SIZE8;
 			break;
-	        case PG_TYPE_DATE:
-		       _type = DBI_TYPE_DATETIME;
-                       _attribs |= DBI_DATETIME_DATE;
-                       break;
-	        case PG_TYPE_TIME:
-                       _type = DBI_TYPE_DATETIME;
-                       _attribs |= DBI_DATETIME_TIME;
-                       break;
-	case PG_TYPE_DATETIME:
-	case PG_TYPE_TIMESTAMP:
+
+        case PG_TYPE_DATE:
+	        _type = DBI_TYPE_DATETIME;
+            _attribs |= DBI_DATETIME_DATE;
+            break;
+        case PG_TYPE_TIME:
+        case PG_TYPE_TIMETZ:
+            _type = DBI_TYPE_DATETIME;
+            _attribs |= DBI_DATETIME_TIME;
+            break;
+        case PG_TYPE_TIMESTAMP:
+        case PG_TYPE_TIMESTAMPTZ:
 			_type = DBI_TYPE_DATETIME;
 			_attribs |= DBI_DATETIME_DATE;
 			_attribs |= DBI_DATETIME_TIME;
 			break;
+
 		case PG_TYPE_NAME:
 		case PG_TYPE_TEXT:
 		case PG_TYPE_CHAR2:
@@ -495,36 +657,38 @@ void _get_field_info(dbi_result_t *result) {
 }
 
 void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned long long rowidx) {
-	int curfield = 0;
+	unsigned int curfield = 0;
 	char *raw = NULL;
-	unsigned long long strsize = 0;
-	unsigned long sizeattrib;
+	size_t strsize = 0;
+	unsigned int sizeattrib;
 	dbi_data_t *data;
+	unsigned char *temp = NULL;
+	size_t unquoted_length;
+
 
 	while (curfield < result->numfields) {
 		raw = PQgetvalue((PGresult *)result->result_handle, rowidx, curfield);
-		strsize = (unsigned long long) PQfmod((PGresult *)result->result_handle, curfield);
 		data = &row->field_values[curfield];
 
 		row->field_sizes[curfield] = 0;
 		/* will be set to strlen later on for strings */
 		
 		if (PQgetisnull((PGresult *)result->result_handle, rowidx, curfield) == 1) {
+		        _set_field_flag( row, curfield, DBI_VALUE_NULL, 1);
 			curfield++;
 			continue;
 		}
 		
 		switch (result->field_types[curfield]) {
 			case DBI_TYPE_INTEGER:
-				sizeattrib = _isolate_attrib(result->field_attribs[curfield], DBI_INTEGER_SIZE1, DBI_INTEGER_SIZE8);
-				switch (sizeattrib) {
+				switch (result->field_attribs[curfield] & DBI_INTEGER_SIZEMASK) {
 					case DBI_INTEGER_SIZE1:
 						data->d_char = (char) atol(raw); break;
 					case DBI_INTEGER_SIZE2:
 						data->d_short = (short) atol(raw); break;
 					case DBI_INTEGER_SIZE3:
 					case DBI_INTEGER_SIZE4:
-						data->d_long = (long) atol(raw); break;
+						data->d_long = (int) atol(raw); break;
 					case DBI_INTEGER_SIZE8:
 						data->d_longlong = (long long) atoll(raw); break; /* hah, wonder if that'll work */
 					default:
@@ -532,8 +696,7 @@ void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned long long rowi
 				}
 				break;
 			case DBI_TYPE_DECIMAL:
-				sizeattrib = _isolate_attrib(result->field_attribs[curfield], DBI_DECIMAL_SIZE4, DBI_DECIMAL_SIZE8);
-				switch (sizeattrib) {
+				switch (result->field_attribs[curfield] & DBI_DECIMAL_SIZEMASK) {
 					case DBI_DECIMAL_SIZE4:
 						data->d_float = (float) strtod(raw, NULL); break;
 					case DBI_DECIMAL_SIZE8:
@@ -543,17 +706,28 @@ void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned long long rowi
 				}
 				break;
 			case DBI_TYPE_STRING:
+			    strsize = (size_t)PQgetlength((PGresult *)result->result_handle, rowidx, curfield);
 				data->d_string = strdup(raw);
 				row->field_sizes[curfield] = strsize;
 				break;
-			case DBI_TYPE_BINARY:
-				row->field_sizes[curfield] = strsize;
-				data->d_string = malloc(strsize);
-				memcpy(data->d_string, raw, strsize);
+			case DBI_TYPE_BINARY:	
+			  temp = PQunescapeBytea((const unsigned char *)raw, &unquoted_length);
+			  if ((data->d_string = malloc(unquoted_length)) == NULL) {
+			    PQfreemem(temp);
+			    break;
+			  }
+			  memmove(data->d_string, temp, unquoted_length);
+			  PQfreemem(temp);
+			  row->field_sizes[curfield] = unquoted_length;
+			  /* todo: is raw ever unescaped binary data? */
+/*        			        strsize = PQgetlength((PGresult *)result->result_handle, rowidx, curfield); */
+/* 				row->field_sizes[curfield] = strsize; */
+/* 				data->d_string = malloc(strsize); */
+/* 				memcpy(data->d_string, raw, strsize); */
 				break;
 				
 			case DBI_TYPE_DATETIME:
-				sizeattrib = _isolate_attrib(result->field_attribs[curfield], DBI_DATETIME_DATE, DBI_DATETIME_TIME);
+				sizeattrib = result->field_attribs[curfield] & (DBI_DATETIME_DATE|DBI_DATETIME_TIME);
 				data->d_datetime = _dbd_parse_datetime(raw, sizeattrib);
 				break;
 				
