@@ -22,7 +22,7 @@
  * Copyright (C) 2002, Markus Hoenicka <mhoenicka@users.sourceforge.net>
  * http://libdbi-drivers.sourceforge.net
  * 
- * $Id: dbd_sqlite.c,v 1.23 2004/01/10 23:15:25 mhoenicka Exp $
+ * $Id: dbd_sqlite.c,v 1.35 2006/10/17 19:18:47 mhoenicka Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -80,7 +80,7 @@ void _translate_sqlite_type(enum enum_field_types fieldtype, unsigned short *typ
 void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned int rowidx);
 int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement);
 char* get_field_type(const char* statement, const char* curr_field_name);
-static unsigned long sqlite_escape_string(char *to, const char *from, unsigned long length);
+static size_t sqlite_escape_string(char *to, const char *from, size_t length);
 int wild_case_compare(const char *str,const char *str_end,
 		      const char *wildstr,const char *wildend,
 		      char escape);
@@ -124,6 +124,8 @@ int _real_dbd_connect(dbi_conn_t *conn, const char* database) {
   const char *dbname;
   const char *dbdir;
 
+  int timeout;
+
   /* initialize error stuff */
   conn->error_number = 0;
   conn->error_message = NULL;
@@ -195,7 +197,16 @@ int _real_dbd_connect(dbi_conn_t *conn, const char* database) {
       conn->current_db = strdup(dbname);
     }
   }
-	
+
+  /* set the SQLite timeout to timeout milliseconds */
+  timeout = dbi_conn_get_option_numeric(conn, "sqlite_timeout");
+
+  if (timeout == -1) {
+    timeout = 0;
+  }
+
+  sqlite_busy_timeout(sqcon, timeout);	
+  
   return 0;
 }
 
@@ -213,16 +224,16 @@ int dbd_disconnect(dbi_conn_t *conn) {
   return 0;
 }
 
-int dbd_fetch_row(dbi_result_t *result, unsigned long long rownum) {
+int dbd_fetch_row(dbi_result_t *result, unsigned long long rowidx) {
   dbi_row_t *row = NULL;
 
-  if (result->result_state == NOTHING_RETURNED) return -1;
+  if (result->result_state == NOTHING_RETURNED) return 0;
 	
   if (result->result_state == ROWS_RETURNED) {
     /* get row here */
     row = _dbd_row_allocate(result->numfields);
-    _get_row_data(result, row, rownum);
-    _dbd_row_finalize(result, row, rownum);
+    _get_row_data(result, row, rowidx);
+    _dbd_row_finalize(result, row, rowidx);
   }
 	
   return 1; /* 0 on error, 1 on successful fetchrow */
@@ -235,8 +246,8 @@ int dbd_free_query(dbi_result_t *result) {
   return 0;
 }
 
-int dbd_goto_row(dbi_result_t *result, unsigned long long row) {
-  result->currowidx = row;
+int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx) {
+  result->currowidx = rowidx;
   return 1;
 }
 
@@ -257,6 +268,37 @@ const char *dbd_get_encoding(dbi_conn_t *conn){
   }
 }
 
+const char* dbd_encoding_to_iana(const char *db_encoding) {
+  /* nothing to translate, return original encoding */
+  return db_encoding;
+}
+
+const char* dbd_encoding_from_iana(const char *iana_encoding) {
+  /* nothing to translate, return original encoding */
+  return iana_encoding;
+}
+
+char *dbd_get_engine_version(dbi_conn_t *conn, char *versionstring) {
+  dbi_result_t *dbi_result;
+  const char *versioninfo = NULL;
+
+  /* initialize return string */
+  *versionstring = '\0';
+
+  dbi_result = dbd_query(conn, "SELECT sqlite_version()");
+
+  if (dbi_result) {
+    if (dbi_result_next_row(dbi_result)) {
+      versioninfo = dbi_result_get_string_idx(dbi_result, 1);
+      strncpy(versionstring, versioninfo, VERSIONSTRING_LENGTH-1);
+      versionstring[VERSIONSTRING_LENGTH-1] = '\0';
+    }
+    dbi_result_free(dbi_result);
+  }
+
+  return versionstring;
+}
+
 dbi_result_t *dbd_list_dbs(dbi_conn_t *conn, const char *pattern) {
   char *sq_errmsg = NULL;
   char old_cwd[_POSIX_PATH_MAX] = "";
@@ -411,13 +453,13 @@ dbi_result_t *dbd_list_tables(dbi_conn_t *conn, const char *db, const char *patt
   return dbd_query(conn, "SELECT tablename FROM libdbi_tablenames ORDER BY tablename");
 }
 
-dbi_result_t *dbd_query_null(dbi_conn_t *conn, const unsigned char *statement, unsigned long st_length) {
+dbi_result_t *dbd_query_null(dbi_conn_t *conn, const unsigned char *statement, size_t st_length) {
 	return NULL;
 }
 
-int dbd_quote_string(dbi_driver_t *driver, const char *orig, char *dest) {
+size_t dbd_quote_string(dbi_driver_t *driver, const char *orig, char *dest) {
   /* foo's -> 'foo\'s' */
-  unsigned int len;
+  size_t len;
 	
   strcpy(dest, "'");
   len = sqlite_escape_string(dest+1, orig, strlen(orig));	
@@ -426,6 +468,32 @@ int dbd_quote_string(dbi_driver_t *driver, const char *orig, char *dest) {
   return len+2;
 }
 
+size_t dbd_conn_quote_string(dbi_conn_t *conn, const char *orig, char *dest) {
+  return dbd_quote_string(conn->driver, orig, dest);
+}
+
+size_t dbd_quote_binary(dbi_conn_t *conn, const unsigned char *orig, size_t from_length, unsigned char **ptr_dest) {
+  unsigned char *temp;
+  size_t len;
+
+  if ((temp = malloc(from_length*2)) == NULL) {
+    return 0;
+  }
+
+  strcpy((char *)temp, "\'");
+  if (from_length) {
+    len = _dbd_encode_binary(orig, from_length, temp+1);
+  }
+  else {
+    len = 0;
+  }
+  strcat((char *)temp, "'");
+
+  *ptr_dest = temp;
+
+  return len+2;
+}
+
 dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
   /* allocate a new dbi_result_t and fill its applicable members:
    * 
@@ -438,7 +506,7 @@ dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
   int numcols;
   char** result_table;
   char* errmsg;
-  unsigned short idx = 0;
+  int idx = 0;
   unsigned short fieldtype;
   unsigned int fieldattribs;
   dbi_error_flag errflag = 0;
@@ -451,16 +519,17 @@ dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
 			       &errmsg);
 
   if (query_res) {
-    _error_handler(conn, errflag);
+    _dbd_internal_error_handler(conn, errmsg, query_res);
     return NULL;
   }
 	
-  result = _dbd_result_create(conn, (void *)result_table, (unsigned long long)numrows, (unsigned long long)sqlite_changes((sqlite*)conn->connection));
+  result = _dbd_result_create(conn, (void *)result_table, numrows, (unsigned long long)sqlite_changes((sqlite*)conn->connection));
 /*   printf("numrows:%d, numcols:%d<<\n", numrows, numcols); */
-  _dbd_result_set_numfields(result, (unsigned short)numcols);
+  _dbd_result_set_numfields(result, numcols);
 
   /* assign types to result */
-  while (idx < (unsigned short)numcols) {
+  while (idx < numcols) {
+/*     printf("idx: %d<< numcols:%d\n", idx, numcols); */
     int type;
     char *item;
     
@@ -517,7 +586,7 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
   char *curr_type;
   char* errmsg;
   int query_res;
-  int table_numrows = 0;
+  int table_numrows = 0; /* int seems ok as sqlite does not use longlongs */
   int table_numcols = 0;
   int type;
   dbi_error_flag errflag = 0;
@@ -552,14 +621,12 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
     }
 
     /* table now points to the table name; find the end of table */
-    item = strchr(table, ' ');
-    if (!item) {
-      strcpy(curr_table, table);
-    }
-    else {
-      strncpy(curr_table, table, item-table);
-      curr_table[item-table] = '\0'; /* terminate just in case */
+    item = table;
+    while (*item && *item != ' ' && *item != ',' && *item != ';') {
+      item++;
     }
+    strncpy(curr_table, table, item-table);
+    curr_table[item-table] = '\0'; /* terminate just in case */
 
     /* for obvious reasons, the internal tables do not contain the
        commands how they were created themselves. We have to use known
@@ -583,7 +650,6 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
   }
 
 /*   printf("curr_table went to %s<<\ncurr_field_name went to %s<<\n", curr_table, curr_field_name); */
-
   /* check for known functions which may appear here instead
      of field names. There is some overlap, i.e. some function work
      both on strings and numbers. These cases would have to be
@@ -647,7 +713,7 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
 					curr_table);
     
     if (query_res || !table_numrows) {
-      _error_handler(conn, errflag);
+      _dbd_internal_error_handler(conn, errmsg, query_res);
 /*       printf("field not found\n"); */
       return 0;
     }
@@ -658,6 +724,8 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
      the table containing the current field */
   /*  parse the sql statement to find the type of the current field */
 /*   printf("table_result_table[3]=%s<<\ncurr_field_name=%s<<\n", table_result_table[3], curr_field_name); */
+
+  /* curr_type will point to an allocated string */
   curr_type = get_field_type(table_result_table[3], curr_field_name);
 
   /* free memory */
@@ -683,7 +751,10 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
 
 /*   printf("field type: %s<<\n", curr_type); */
   if (strstr(curr_type, "BLOB")
-      || strstr(curr_type, "CHAR(") /* note the opening bracket */
+      || strstr(curr_type, "BYTEA")) {
+    type = FIELD_TYPE_BLOB;
+  }
+  else if (strstr(curr_type, "CHAR(") /* note the opening bracket */
       || strstr(curr_type, "CLOB")
       || strstr(curr_type, "TEXT") /* also catches TINYTEXT */
       || strstr(curr_type, "VARCHAR")
@@ -741,9 +812,9 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
   else {
     type = FIELD_TYPE_STRING; /* most reasonable default */
   }
-  
   free(curr_type);
 
+/*   printf("type went to %d<<\n", type); */
   return type;
 }
 
@@ -816,7 +887,7 @@ char* get_field_type(const char* statement, const char* curr_field_name) {
   return curr_type;
 }
 
-char *dbd_select_db(dbi_conn_t *conn, const char *db) {
+const char *dbd_select_db(dbi_conn_t *conn, const char *db) {
   /*
     sqlite does not separate connecting to a database server and using
     or opening a database. If we want to switch to a different database,
@@ -836,7 +907,7 @@ char *dbd_select_db(dbi_conn_t *conn, const char *db) {
     return NULL;
   }
 
-  return (char *)db;
+  return db;
 }
 
 int dbd_geterror(dbi_conn_t *conn, int *errno, char **errstr) {
@@ -849,7 +920,7 @@ int dbd_geterror(dbi_conn_t *conn, int *errno, char **errstr) {
     result++;
   }
   if (conn->error_message) {
-    *errstr = conn->error_message;
+    *errstr = strdup(conn->error_message);
     result += 2;
   }
   return result;
@@ -956,9 +1027,9 @@ void _translate_sqlite_type(enum enum_field_types fieldtype, unsigned short *typ
 void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned int rowidx) {
   char **result_table = result->result_handle;
   
-  int curfield = 0;
+  unsigned int curfield = 0;
   char *raw = NULL;
-  unsigned long sizeattrib;
+  unsigned int sizeattrib;
   dbi_data_t *data;
 
   while (curfield < result->numfields) {
@@ -985,7 +1056,7 @@ void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned int rowidx) {
 	data->d_short = (short) atol(raw); break;
       case DBI_INTEGER_SIZE3:
       case DBI_INTEGER_SIZE4:
-	data->d_long = (long) atol(raw); break;
+	data->d_long = (int) atol(raw); break;
       case DBI_INTEGER_SIZE8:
 	data->d_longlong = (long long) atoll(raw); break; /* hah, wonder if that'll work */
       default:
@@ -1008,9 +1079,8 @@ void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned int rowidx) {
       row->field_sizes[curfield] = strlen(raw);
       break;
     case DBI_TYPE_BINARY:
-      /* treat as string */
       data->d_string = strdup(raw);
-      row->field_sizes[curfield] = strlen(raw);
+      row->field_sizes[curfield] = _dbd_decode_binary(data->d_string, data->d_string);
       break;
     case DBI_TYPE_DATETIME:
       sizeattrib = _isolate_attrib(result->field_attribs[curfield], DBI_DATETIME_DATE, DBI_DATETIME_TIME);
@@ -1118,8 +1188,8 @@ int wild_case_compare(const char *str,const char *str_end,
 
 /* this function is stolen from MySQL. The quoting was changed to the
  SQL standard, i.e. single and double quotes are escaped by doubling,
- not by a backslash */
-static unsigned long sqlite_escape_string(char *to, const char *from, unsigned long length)
+ not by a backslash. Newlines and carriage returns are left alone */
+static size_t sqlite_escape_string(char *to, const char *from, size_t length)
 {
   const char *to_start=to;
   const char *end;
@@ -1131,26 +1201,10 @@ static unsigned long sqlite_escape_string(char *to, const char *from, unsigned l
 	*to++= '\\';
 	*to++= '0';
 	break;
-      case '\n':				/* Must be escaped for logs */
-	*to++= '\\';
-	*to++= 'n';
-	break;
-      case '\r':
-	*to++= '\\';
-	*to++= 'r';
-	break;
-/*       case '\\': */
-/* 	*to++= '\\'; */
-/* 	*to++= '\\'; */
-/* 	break; */
       case '\'':
 	*to++= '\''; /* double single quote */
 	*to++= '\'';
 	break;
-/*       case '"':				/\* Better safe than sorry *\/ */
-/* 	*to++= '"'; /\* double double quote *\/ */
-/* 	*to++= '"'; */
-/* 	break; */
       case '\032':			/* This gives problems on Win32 */
 	*to++= '\\';
 	*to++= 'Z';
@@ -1160,5 +1214,6 @@ static unsigned long sqlite_escape_string(char *to, const char *from, unsigned l
       }
     }
   *to=0;
-  return (unsigned long) (to-to_start);
+  return (size_t) (to-to_start);
 }
+
